# -*- coding: utf-8 -*-

"""Main Controller"""
from tg import expose, flash, require, url, request, redirect, tmpl_context
from pylons.i18n import ugettext as _
from repoze.what import predicates

from luci.lib.base import BaseController

from luci.controllers.error import ErrorController
from luci.controllers.cluster import ClusterController
from luci.controllers.async import AsyncController

import luci.widget_validators.validate_cluster_prop as vcp
from luci.lib.db_helpers import db_remove_cluster

__all__ = ['RootController']


# Imports into module's namespace.


class RootController(BaseController):
    """
    The root controller for the luci application.

    All the other controllers and WSGI applications should be mounted on this
    controller. For example::

        panel = ControlPanelController()
        another_app = AnotherWSGIApplication()

    Keep in mind that WSGI applications shouldn't be mounted directly: They
    must be wrapped around with :class:`tg.controllers.WSGIAppController`.

    """

    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        tmpl_context.title_list = []
        tmpl_context.already_used = dict()
        tmpl_context.already_used.setdefault(False)
        return BaseController.__call__(self, environ, start_response)

    # SUBCONTROLLERS

    error = ErrorController()
    cluster = ClusterController()
    async = AsyncController()

    # METHODS OF THE ROOT CONTROLLER
    
    @expose('luci.templates.index')
    def index(self):
        """Handle the front-page."""
        tmpl_context.show_sidebar = True
        redirect('/homebase/')
        return dict(page='index')

    @expose('luci.templates.about')
    def about(self):
        """Handle the 'about' page."""
        return dict(page='about')

    @expose('luci.templates.homebase')
    @require(predicates.not_anonymous())
    def homebase(self, homebasepage='homebasepage', **args):
        tmpl_context.show_sidebar = True
        tmpl_context.homebase = True
        return dict(page='homebase', homebasepage=homebasepage, args=args,
                    base_url='/cluster')

    @expose('luci.templates.homebase')
    @require(predicates.not_anonymous())
    def add_existing_cmd(self, **args):
        vret = vcp.validate_add_existing(**args)
        if vret[0] is True:
            msgs = vret[1].get('flash')
            if msgs and len(msgs) > 0:
                flash(', '.join(msgs))
        else:
            msgs = vret[1].get('errors')
            if msgs and len(msgs) > 0:
                flash(', '.join(msgs), status="error")
        redirect('/homebase')

    @expose('luci.templates.homebase')
    @require(predicates.has_permission('manage'))
    def manage_remove_cmd(self, **args):
        errors = []
        success = []
        for k, v in args.items():
            if db_remove_cluster(k) is not True:
                errors.append(k)
            else:
                success.append(k)
        if len(errors) != 0:
            flash(_("Unable to remove cluster %s from the luci interface")
                % (', '.join(errors)), status="error")
        if len(success) != 0:
            flash(_("Removed cluster %s from the luci interface")
                % (', '.join(errors)), status="info")
        redirect('/homebase')

    @expose('luci.templates.cluster')
    def clusters(self, **args):
      redirect('/cluster')

    @expose('luci.templates.login')
    def login(self, came_from=url('/')):
        """Start the user login."""
        login_counter = request.environ['repoze.who.logins']
        if login_counter > 0:
            flash(_('Wrong credentials'), 'warning')
        return dict(page='login', login_counter=str(login_counter),
                    came_from=came_from)

    @expose()
    def post_login(self, came_from=url('/')):
        """
        Redirect the user to the initially requested page on successful
        authentication or redirect her back to the login page if login failed.

        """
        if not request.identity:
            login_counter = request.environ['repoze.who.logins'] + 1
            redirect(url('/login', came_from=came_from, __logins=login_counter))
        userid = request.identity['repoze.who.userid']
        flash(_('Welcome back, %s!') % userid)
        redirect(came_from)

    @expose()
    def post_logout(self, came_from=url('/')):
        """
        Redirect the user to the initially requested page on logout and say
        goodbye as well.

        """
        flash(_('We hope to see you soon!'))
        redirect(came_from)
