# Copyright (C) 2009-2010 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

# -*- coding: utf-8 -*-

"""Constants and functions regarding the (logical and URL) scheme of Luci.

Here is also a place, where strings forming certain parts of URL across
Luci can be adjusted.

"""

from luci.lib.helpers import singleton

__all__ = ['LuciScheme']


class Scheme():
    """Base abstract class for scheme structure."""

    def __getattr__(self, name):
        if name=='mounts':
            return self.getMounts()
        elif name=='getMounts':
            raise RuntimeError, \
                  '\'getMounts\' method has to be defined in derived class'


@singleton
class LuciScheme(Scheme):
    """Encapsulation of logical and URL scheme of Luci.

    The root of recursively nested classes (`singletons') deriving from
    the same base class as this one.

    """

    # Static URL prefix (excluding host part and slashes both at the beginning
    # and at the end). E.g. 'luci' for Luci application set to be running at
    # <http://example.com/luci/> (notice: appropriate changes in controllers
    # schema have to be made to reach this behavior).
    APP_PREFIX = u''

    # First level of the scheme.

    CLUSTERS = u'clusters'
    GLOBAL_RES = u'global-resources'
    STORAGE = u'storage'
    USERS = u'users'
    SYSTEMS = u'systems'

    @classmethod
    def getMounts(cls):
        """Defined as a method to avoid circular dependency."""

        from luci.controllers.cluster import ClusterController, \
                                             ClusterCmdController
        from luci.controllers.global_res import GlobalResController, \
                                                GlobalResCmdController
        return {
            cls.CLUSTERS:   (ClusterController, ClusterCmdController),
            cls.GLOBAL_RES: (GlobalResController, GlobalResCmdController)
        }

    # Second level of the scheme -- `Clusters'.

    @singleton
    class Cluster(Scheme):
        NODES = u'nodes'
        SERVICES = u'services'
        FAILOVERS = u'failovers'
        FENCES = u'fences'

        @classmethod
        def getMounts(cls):
            """Defined as a method to avoid circular dependency."""
            from cluster_part.node import NodeController, NodeCmdController
            from cluster_part.service import ServiceController, ServiceCmdController
            from cluster_part.failover import FailoverController, FailoverCmdController
            from cluster_part.fence import FenceController, FenceCmdController
            
            return {
                cls.NODES:     (NodeController, NodeCmdController),
                cls.SERVICES:  (ServiceController, ServiceCmdController),
                cls.FAILOVERS: (FailoverController, FailoverCmdController),
                cls.FENCES:    (FenceController, FenceCmdController)
            }


    #---------------------------------------------------------------------------
    # Part related to `base controller' - `commands controller' correspondence.

    """Suffix which makes identifier of base subcontroller unique.

    Such unique identifier can be then used as an identifier of `commands
    subcontroller'.

    """
    CTRL_CMD_SUFFIX = u'_cmd'

    @classmethod
    def base2CmdCtrl(cls, base):
        """Convert `base controller' -- `command subcontroller' name."""
        return base + cls.CTRL_CMD_SUFFIX

    @classmethod
    def cmd2BaseCtrl(cls, cmd):
        """Convert `command subcontroller' -- `base controller' name."""
        return cmd[:cmd.rfind(cls.CTRL_CMD_SUFFIX)]

